import utils
import argparse
import pygments
import pygments.lexers
import pygments.formatters

class StatNode:
    def __init__(self,name,node_type,parent=None):
        self.name      = name
        self.node_type = node_type
        self.parent    = parent
        self.loc       = 0
        self.path      = self._path()
        
    def incr_stat(self):
        node = self
        while node:
            node.loc += 1
            node      = node.parent

    def has_key(self,name):
        return any(x.has_key(name) for x in self.get_catalog())

    def get(self,name):
        """
        Used by get_node. It can't be renamed to _get because most of time get_node operates on dicts.
        This is to fake a dictionary-like interface.
        """
        for x in self.get_catalog():
            # x can be a dict or a CodeStatNode
            if x.has_key(name):
                return x.get(name)

    def get_node(self,path):
        """
        self.get_node("/some/path/to/node") => returns the "node" node.
        special case : if path = "..", return the parent.
        This makes the writing of do_cd more elegant
        """
        utils.debug("[DEBUG StatNode.get_node] %s.get_node(%s)" % (self.path,path))
        if not path:
            return self
        node       = self
        split_path = path.split("/")
        next_node_name = split_path[0]
        if next_node_name == "..":
            next_node = node.parent
        else:
            for entry in self.get_catalog():
                # entry can be a dict or a CodeStatNode
                if entry.has_key(next_node_name):
                    child = entry
                    break
            else:
                return
            next_node = child.get(next_node_name)
        if not next_node :
            return 
        return next_node.get_node("/".join(split_path[1:]))

    def _path(self):
        node    = self
        parents = [node]
        # first, get the list of all the parents
        while node:
            parent = node.parent
            if not parent :
                break
            parents.append(parent)
            node = parent
        # then traverse them in reverse order
        return "/".join([parent.name for parent in parents[::-1]])
        
    def _display_listing(self,nodes,max_length):
        utils.debug("[DEBUG _display_listing] max_length",max_length)
        if type(nodes) != list :
            nodes = [nodes]
        for node in nodes :
            loc = node.loc
            utils.debug("[DEBUG _display_listing] loc of current node is",loc)
            print "\t".join((node.node_type.ljust(10),
                             node.name.ljust(max_length),
                             str(node.loc).rjust(9)
                            ))
    
    def cat(self,style=None):
        code = "\n".join(self.get_code())
        if style :
            print pygments.highlight(code,pygments.lexers.PythonLexer(),pygments.formatters.Terminal256Formatter(style=style,bg="dark"))
        else :
            print pygments.highlight(code,pygments.lexers.PythonLexer(),pygments.formatters.TerminalFormatter(colorscheme=utils.CAT_COLORS,bg="dark"))


    @utils.hide_problem
    def sort(self,nodeslist,sort='',reverse=None):
        if not sort:
            return nodeslist
        accessfunc = None
        if sort in ("size","loc"):
            accessfunc = lambda x:x.loc
        elif sort == "name":
            accessfunc = lambda x:x.name
        return sorted(nodeslist,key=accessfunc,reverse=reverse)

class PackageStatNode(StatNode):
    def __init__(self,name,parent):
        StatNode.__init__(self,name,node_type="package",parent=parent)
        self.packages = {}
        self.modules  = {}

    def ls(self,sort='',reverse=None):
        utils.debug("[DEBUG PackageStatNode.ls] packages",self.packages)
        packages_names = [package.name for package in self.packages.values()]
        modules_names  = [module.name for module in self.modules.values()]
        total          = modules_names+packages_names
        if total :
            max_length = len(max(total,key=lambda x:len(x)))
        else:
            max_length = 10
        packages = self.sort(self.packages.values(),sort,reverse)
        modules  = self.sort(self.modules.values(),sort,reverse)
        self._display_listing(packages,max_length)
        self._display_listing(modules,max_length)
        print "Total : %s lines of code in %s modules and %s packages" % (self.loc,len(modules_names),len(packages_names))        
        
    def cat(self,style=None):
        print "<PackageStatNode.cat not implemented>"

    def get_catalog(self):
        return [self.packages,self.modules]

    def get_code(self):
        return ""
        
class ModuleStatNode(StatNode):
    """
    A StatNode that represents anything that can be defined in a module : a function, a class, and the entire module itself.
    """
    def __init__(self,name,parent,node_type,depth=0):
        StatNode.__init__(self,name,parent=parent,node_type=node_type)
        self.depth     = depth
        self.classes   = {}
        self.functions = {}
        self.code      = CodeStatNode(parent=self)
        
    def get_catalog(self):
        return [self.classes,self.functions,self.code]
                
    def ls(self,sort="",reverse=None):
        # 1) Get the names of all the classes
        # 2) Get the names of all the functions
        # 3) Get max_name
        # 4) print the names, adjust on max_name, sort
        utils.debug("[DEBUG ModuleStatNode.ls] %s.ls()" % self.path)
        classes_names   = [klass.name    for klass    in self.classes.values()   ]
        functions_names = [function.name for function in self.functions.values() ]
        utils.debug("[DEBUG ModuleStatNode.ls] classes_names"   ,classes_names)
        utils.debug("[DEBUG ModuleStatNode.ls] functions_names" ,functions_names)
        total = classes_names + functions_names
        if total :
            max_name = len(max(total,key=lambda x:len(x)))
        else:
            max_name = 10

        self._display_listing(self.sort(self.classes.values()   ,sort,reverse) , max_name)
        self._display_listing(self.sort(self.functions.values() ,sort,reverse) , max_name)
        self._display_listing(self.code                                        , max_name)
        print "Total : %s loc in %s classes and %s functions" % (self.loc,len(self.classes),len(self.functions))

        
    def get_code(self):
        yield self.code.source        
        for node in self.get_catalog():
            for child in node.values():
                for code in child.get_code():
                    yield code

        
class CodeStatNode(StatNode):
    """
    A class that represents actual source code.
    It defines an extra attribute to store source code.
    """
    def __init__(self,parent,depth=0):
        StatNode.__init__(self,name="code",node_type="code",parent=parent)
        self.depth     = depth
        self.classes   = {}
        self.functions = {}
        self.source    = ""

    def get_catalog(self):
        return [self.classes,self.functions]

    def values(self):
        return []

    def keys(self):
        return []
    
    def get_code(self):
        return self.source

    def ls(self):
        print "<CodeStatNode.ls> Not Implemented"
